grammar DuraC;

options {
  output=AST;
}


tokens {
  PROGRAM;
  MODULE;
  MODULE_NAME;
  LOAD;
  IMPORT;
  EXPORT;
  DESCRIPTION;
  PREAMBLE;
  DEFINITIONS;
  
  SCHEMA;
  TUPLE_CONSTRAINTS;
  RELATION_CONSTRAINTS;
  
  TYPEDEF;
  TYPE;
  CONSTDEF;
  MODIFIER;
  MODIFIERS;
  EVENT;
  STATEFUL_OBJECT;
  ACTION;

  EVENT_QUERY;
  STATE_QUERY;
  
  FOR;
  ACTION_EXECUTION;
  EXECUTION_CONSTRAINTS;
  SUCCEEDS_ON;
  FAILS_ON;
  
  QUERY;
  HEAD;
  BODY;
   
  IDENTIFIERS;

  SUPPLEMENT;
  AGGREGATION;
  GROUPING;
  CONDITION;
  
  DETECT;
  DO;
  LET;
  
  CONJUNCTION;
  DISJUNCTION;
  CONCURRENT;
  NEGATION;
  EXISTS;

  ARITHMETIC_OP;
  INTERVAL_OP;
  AGGREGATION_OP;
  RELATIVE_TIMER_OP;
  CLOCK_FUNCTION;
  
  TIME_RELATION;
  ARITHMETIC_RELATION;
  UNIFICATION;
  
  PATH;
  TERM;
  VARIABLE;
  TYPED_VARIABLE;
  IDENTIFIER;
  CONSTANT;
  CONSTRAINT;

  BASIC_TYPE;
  COMPOSITE_TYPE;
  TIME_UNIT;
  DURATION;
  BOOLEAN;
  
  ANNOTATION;
}

@header {
package de.lmu.ifi.pms.cep.dura.compiler.antlr;

import de.lmu.ifi.pms.cep.dura.compiler.exceptions.DuraParseException;
}

@lexer::header {
package de.lmu.ifi.pms.cep.dura.compiler.antlr;
}

@members {
	public void displayRecognitionError(String[] tokenNames, RecognitionException e) {
	
		String hdr = getErrorHeader(e);
		String msg = getErrorMessage(e, tokenNames);
		
		throw new DuraParseException(msg, getSourceName(), e.line);
	}
}

/* DURA Core*/

program
  :  preamble (def+=eventDefinition | def+=stateDefinition | def+=actionDefinition | def+=reactiveRule)*
  ->  ^(PROGRAM ^(MODULE preamble ^(DEFINITIONS $def*)))
  ;

preamble
  : pckg description load* imprt* export* (constDefinition | typeDefinition)*
  ->  ^(PREAMBLE pckg description load* imprt* export* constDefinition* typeDefinition*)
  ;

pckg
  : ('MODULE' path)?
  ->  ^(MODULE_NAME path?)
  ;
  
description
  : ('DESCRIPTION' STRING)?
  ->  ^(DESCRIPTION STRING?)
  ;
  
imprt
  : ('IMPORT' path)
  -> ^(IMPORT path)
  ;

load
	:	('LOAD' path)
	-> ^(LOAD path)
	;
	
export
	:	('EXPORT' path)
	-> ^(EXPORT path)
	;

/** 
 * constraints: 
 *   all schemaDefinition labels are pairwise distinct
 *   type name ID is unique
 *   no cyclic definitions
 */
typeDefinition 
  : visibilityModifier? 'TYPE' ID 'IS' basicType typeSupplement 'END'
  ->  ^(TYPEDEF[$ID] ^(MODIFIERS visibilityModifier?) ^(TYPE basicType typeSupplement))
  | visibilityModifier? 'TYPE' ID 'IS' schemaDefinition typeSupplement 'END'
  ->  ^(TYPEDEF[$ID] ^(MODIFIERS visibilityModifier?) ^(TYPE schemaDefinition typeSupplement))
  | visibilityModifier? 'TYPE' ID 'IS' '{' schemaDefinition (',' schemaDefinition)* '}' typeSupplement 'END'  
  ->  ^(TYPEDEF[$ID] ^(MODIFIERS visibilityModifier?) ^(TYPE schemaDefinition+ typeSupplement))
  ;

/** constraint: const name ID is unique */
constDefinition
  : visibilityModifier? 'CONST' ID 'IS' constTerm 'END'      ->  ^(CONSTDEF[$ID] ^(MODIFIERS visibilityModifier?) constTerm)
  ;


/**
 * constraints:
 *   event definition schemaDefinition needs to be unique
 *   schema of derived events need to comply with schmaDenfinition
 */
eventDefinition
  : modifiers 'EVENT' schemaDefinition schemaSupplement ('WITH' eventSpecification*)? 'END'
  ->  ^(EVENT modifiers ^(SCHEMA schemaDefinition schemaSupplement) eventSpecification*)
  ;

/** constraint: stateful object definition schemaDefinition is unique */
stateDefinition
  : modifiers 'STATEFUL OBJECT' schemaDefinition schemaSupplement 'END'
  ->  ^(STATEFUL_OBJECT modifiers ^(SCHEMA schemaDefinition schemaSupplement))
  ;

/** constraint: action definition schemaDefinition is unique */
actionDefinition
  : modifiers 'ACTION' schemaDefinition schemaSupplement 'END'
  ->  ^(ACTION modifiers ^(SCHEMA schemaDefinition schemaSupplement))
  ;

/** constraint: query is range restricted with respect to term */
eventSpecification
  : 'DETECT' term 'ON' eventQuery 'END'
  ->  ^(DETECT ^(HEAD term) ^(BODY eventQuery))
  ;

/** constraint: action is range restricted with respect to query */
reactiveRule
  : 'ON' eventQuery 'DO' action 'END'
  ->  ^(DO ^(BODY eventQuery) ^(HEAD action))
  ;
  

/** constaint: at least one event query in every disjunct */
eventQuery
  : 'and' '{' flatEventQuery (','? flatEventQuery)* ','? '}' querySupplement ->  ^(CONJUNCTION flatEventQuery+ querySupplement)
  | 'or' '{' flatEventQuery (','? flatEventQuery)* ','? '}' querySupplement  ->  ^(DISJUNCTION flatEventQuery+ querySupplement)
  | flatEventQuery                                                          ->  ^(CONJUNCTION flatEventQuery ^(SUPPLEMENT))
  ;

flatEventQuery
  : 'not' '{' atomicEventQuery '}' negationSupplement  ->  ^(NEGATION atomicEventQuery negationSupplement)
  | 'not' atomicEventQuery ->  ^(NEGATION atomicEventQuery SUPPLEMENT)
  | atomicEventQuery
  ;

atomicEventQuery
  : i='event' ID ':' qualifiedTerm querySupplement ->  ^(EVENT_QUERY ^(IDENTIFIER[$i] ID) ^(QUERY qualifiedTerm) querySupplement)
  | i='state' ID ':' qualifiedTerm querySupplement ->  ^(STATE_QUERY ^(IDENTIFIER[$i] ID) ^(QUERY qualifiedTerm) querySupplement) 
  ;

action
  : 'concurrent' '{' atomicAction (','? atomicAction)* ','? '}' ->  ^(CONCURRENT atomicAction+)
  | atomicAction                                          ->  ^(CONCURRENT atomicAction+)
  ;

/** constraint: action term is actually defined */  
atomicAction
  : 'action' ID? ':' qualifiedTerm ->  ^(ACTION_EXECUTION qualifiedTerm)
  ;


querySupplement
  : (supplement+=where | supplement+=let | supplement+=grouping)*
  ->  ^(SUPPLEMENT $supplement*)
  ;

negationSupplement
  : where?
  ->  ^(SUPPLEMENT where?)
  ;

actionSupplement
  : where*           // where is too generic here, only temporal relation can be stated in the where part of an expression. var X - var Y <= 0 may be kind of a temporal relation for event queries, but not for action definitions
  ->  ^(SUPPLEMENT where*)
  ;

typeSupplement
  : tupleConstraints?
  -> ^(SUPPLEMENT tupleConstraints?)
  ;

schemaSupplement
  : tupleConstraints? relationConstraints?
  ->  ^(SUPPLEMENT tupleConstraints? relationConstraints?)
  ;

let
  : 'let' '{' unification (','? unification)* ','? '}'
  ->  ^(LET unification+)
  ;

unification
  : dataVariable '=' expr
  -> ^(UNIFICATION dataVariable expr)
  | dataVariable '=' relativeTimerOp '(' timeInterval ',' durationValue ')'  
  -> ^(UNIFICATION dataVariable ^(relativeTimerOp timeInterval durationValue))
  ;

grouping
  : 'group by' '{' binding (',' binding)* '}'
  ->  ^(AGGREGATION ^(GROUPING binding+))
  
  | 'group by' '{' binding (',' binding)* '}' 'aggregate' '{' aggregation (',' aggregation)* '}'
  ->  ^(AGGREGATION ^(GROUPING binding+) ^(AGGREGATION aggregation+))
  ;

binding
  : dataVariable
  | identifier
  ;

aggregation
  : dataVariable '=' aggregationOp '(' binding ')'
  -> ^(UNIFICATION dataVariable ^(aggregationOp binding))
  ;

/** constraint: no variables occur in mathFormula */    
tupleConstraints
  : 'where' '{' atomicMathFormula (',' atomicMathFormula)* '}'
  ->  ^(TUPLE_CONSTRAINTS atomicMathFormula+)
  ;

relationConstraints
  : 'constraints' '{' atomicRelationConstraint (',' atomicRelationConstraint)* '}' 
  -> ^(RELATION_CONSTRAINTS atomicRelationConstraint+)
  ;

atomicRelationConstraint
  : (c='bound' | c='key' | c='primary-key') '(' path ')'  ->  ^(CONSTRAINT[$c] path)
  | path c='references' path '{' path '}'                 ->  ^(CONSTRAINT[$c] path+) 
  | c='default' '(' path '=' clockFunction ')'            ->  ^(CONSTRAINT[$c] ^(UNIFICATION path clockFunction))
  ;

where
  : 'where' 'and' '{' conditionFormula (',' conditionFormula)*  '}'  -> ^(CONDITION ^(CONJUNCTION conditionFormula+))
  | 'where' 'or' '{' conditionFormula (',' conditionFormula)*  '}'   -> ^(CONDITION ^(DISJUNCTION conditionFormula+))  
  | 'where' 'not' '{' conditionFormula (',' conditionFormula)*  '}'  -> ^(CONDITION ^(NEGATION conditionFormula+))  
  | 'where' '{' conditionFormula (',' conditionFormula)*  '}'        -> ^(CONDITION ^(CONJUNCTION conditionFormula+))
  ;

conditionFormula
  : 'and' '{' conditionFormula (',' conditionFormula)* '}'	->  ^(CONJUNCTION conditionFormula+)
  | 'or' '{'conditionFormula (',' conditionFormula)* '}'    ->  ^(DISJUNCTION conditionFormula+)
  | 'not' '{'conditionFormula (',' conditionFormula)* '}'   ->  ^(NEGATION conditionFormula+)
  | atomicConditionFormula
  ;
  
atomicConditionFormula
  :  atomicIntervalFormula
  |	 atomicMathFormula
  ;

/** constraints: 
  *   either both or none of the expressions is of type duration 
  */
atomicMathFormula 
	:	expr arithmeticRelation expr ->  ^(arithmeticRelation expr expr)
	;

atomicIntervalFormula
  :	'{' timeInterval ',' timeInterval 
  	(	                    '}' op='apart-by' durationValue
  	|	(',' timeInterval)* '}' op='within' durationValue
  	)
  ->  ^(TIME_RELATION[$op] ^(IDENTIFIERS timeInterval+) durationValue)
  
  | timeInterval
  	(	intervalRelation timeInterval	->	^(intervalRelation timeInterval timeInterval)
  	|	timepointRelation timePoint	  ->	^(timepointRelation timeInterval timePoint)		
  	)
  ;

/** constraint: variable is of type timeinterval */
timeInterval
  : dataVariable
  |	path
  | identifier
  | relativeTimerOp '(' timeInterval ',' durationValue ')'  ->  ^(relativeTimerOp timeInterval durationValue)
  ;
  
durationValue
	:	duration
	| constant
	;
  
/** constraint: variable is of type timepoint */
timePoint
  : dataVariable
  | path
  | intervalOp '(' timeInterval ')'  ->  ^(intervalOp timeInterval) 
  ;

/** constraint: all timeunits are are pairwise distinct */
duration
  : time+
  ->  ^(DURATION time+)
  ;

time
  : NUMBER timeUnit
  ->  ^(timeUnit NUMBER)
  ;
  
bool
  : (v='true' | v='false')
  ->  BOOLEAN[$v]
  ;

/* XCerpt- */

/** 
 * constraints: 
 *   all term labels are pairwise distinct
 *   no cyclic definitions
 */
term
  : termLeaf                        ->  termLeaf
  | label '{' '}'                   ->  label
  | label '{' term (',' term)* '}' ->  ^(label term+)
  ;

qualifiedTerm
	:	termLeaf                      ->  termLeaf
	| path '{' '}'									->	path
	| path '{' term (',' term)* '}'	->	^(path term+)
	;

/** constraints: 
 *    variables are unified with basic types
 *    constants are unified with basic types
 *    identifiers are unified with identifier types
 */ 
termLeaf
  : dataVariable
  | clockFunction
  | identifier
  | constant
  | duration
  | bool
  | path
  | STRING
  | NUMBER
  ;

/**
 * constraints:
 *   all constTerm labels are pairwise distinct
 *   no cyclic definitions
 */
constTerm
  : STRING
  | NUMBER
  | constant
  | duration
  | bool
  | label '{' '}'                             ->  label
  | label '{' constTerm (',' constTerm)* '}'  ->  ^(label constTerm+)
  ;


/** 
 * constraints: 
 *   all schemaDefinition labels are pairwise distinct
 *   no cyclic definitions
 */
schemaDefinition
  : label '{' '}'                                           ->  label
  | label '{' basicType '}'                                 ->  ^(label basicType)
  | label '{' compositeType '}'                             ->  ^(label compositeType)
  | label '{' schemaDefinition (',' schemaDefinition)* '}'  ->  ^(label schemaDefinition+)
  ;
  
qualifiedSchemaDefinition
  : path '{' '}'                                           ->  path
  | path '{' basicType '}'                                 ->  ^(path basicType)
  | path '{' compositeType '}'                             ->  ^(path compositeType)
  | path '{' schemaDefinition (',' schemaDefinition)* '}'  ->  ^(path schemaDefinition+)
  ;

/** constraint: no duplicate specification of modifiers */
modifiers
  : modifier*
  ->  ^(MODIFIERS modifier*)
  ;

modifier
  : externalModifier
  | visibilityModifier
	| dynamabilityModifier
  ;

externalModifier
	: ('input' | 'output' | 'log' )
	-> MODIFIER[$externalModifier.text]
  ;
  
visibilityModifier
  : ('public' | 'private')
  -> MODIFIER[$visibilityModifier.text]
  ;
  
dynamabilityModifier
  : 'static'
  -> MODIFIER[$dynamabilityModifier.text]
  ;

expr  
  : mathExpr
  | identifier
  ;
    
mathExpr
  : (multExpr -> multExpr)
    ( (op='+' | op='-') multExpr
      ->  ^(ARITHMETIC_OP[$op] $mathExpr multExpr)
    )*
  ;
  
multExpr
  : (powExpr -> powExpr)
    ( (op='*' | op='/') powExpr
      ->  ^(ARITHMETIC_OP[$op] $multExpr powExpr)
    )*
  ;

powExpr
  : (atom -> atom)
    ( op='^' atom
      ->  ^(ARITHMETIC_OP[$op] $powExpr atom)
    )*
  ;

/**
 * constraints:
 *   constants used in aritmetic expressions need to be a number
 */
atom
  : '(' mathExpr ')' 								
  ->  mathExpr
  | (op='least' | op='greatest') '(' mathExpr (',' mathExpr)* ')'  
  ->  ^(ARITHMETIC_OP[$op] mathExpr+)
  | intervalOp '(' timeInterval ')'	
  ->  ^(intervalOp timeInterval)
  | clockFunction
  | dataVariable
  | constant
  | bool
  | op='-' atom
  -> ^(ARITHMETIC_OP[$op] atom)
  | path
  | duration
  | NUMBER
  | STRING
  ;

path
  : label ('.' label)*  ->  PATH[$path.text]
  ;

clockFunction
  : 'system' '.' (fn='now' | fn='tuple_id_seq') '(' ')'  ->  CLOCK_FUNCTION[$fn]
  ;

label
  : aggregationOp 
  ->  TERM[$label.text] 
  | intervalOp        
  ->  TERM[$label.text]
  | ('action' | 'event' | 'state' | 'initiated' | 'failed' | 'succeeded' | 'create' | 'terminate' | 'update' | 'created' | 'terminated' | 'updated')
  ->  TERM[$label.text]
  | ('DESCRIPTION' | 'TYPE' | 'EVENT' | 'ACTION' | 'END')
  ->  TERM[$label.text] 
  | ID
  ->  TERM[$label.text]
  | ID d='$' ('initiated' | 'failed' | 'succeeded'| 'created' | 'terminated' | 'updated')
  -> TERM[$label.text]
  | ID d='$' ('create' | 'terminate' | 'update') ( '$' ('failed' | 'initiated' | 'succeeded'))?
   -> TERM[$label.text]
  ;


type
  : basicType
  | compositeType
  ;

basicType 
  : (t='string' | t='int' | t='long' | t='double' | t='float' | t='boolean' | t='identifier' | t='timestamp' | t='duration')  
  ->  BASIC_TYPE[$t]
  ;

/** constraint: composite type ID is actually defined */
compositeType 
  : path
  ->  COMPOSITE_TYPE[$path.text]
  ;

identifier
  : (i='event' | i='state' | i='action') ID
  ->  ^(IDENTIFIER[$i] ID)
  ;

dataVariable
  : 'var' ID
  ->  ^(VARIABLE ID)
  ;

/*
typedDataVariable
  : 'var' type ID
  ->  ^(TYPED_VARIABLE type ID)
  ;
*/

/** constraint: constant ID is actually defined */
constant
  : 'const' path
  ->  ^(CONSTANT path)
  ;

intervalRelation
  : (rel='before' | rel='contains' | rel='overlaps' | rel='after' | rel='during' | rel='overlapped-by'
      | rel='starts' | rel='finishes' | rel='meets' | rel='started-by' | rel='finished-by'
      | rel='met-by' | rel='equals' | rel='while' | rel='valid-during')
  ->  TIME_RELATION[$rel]
  ;

timepointRelation
	:	rel='valid-at'	-> TIME_RELATION[$rel]
	;

relativeTimerOp 
  : (op='extend' | op='shorten' | op='extend-begin' | op='shorten-begin' | op='shift-forward'
      | op='shift-backward' | op='from-end' | op='from-end-backward' | op='from-start'
      | op='from-start-backward' | op='from-begin' |  op='from-begin-backward')
  ->  RELATIVE_TIMER_OP[$op]
  ;

intervalOp 
  : (op='begin' | op='end')
  ->  INTERVAL_OP[$op]
  ;

aggregationOp 
  : (op='max' | op='min' | op='sum' | op='avg' | op='count')
  ->  AGGREGATION_OP[$op]
  ;

timeUnit
  : (unit='day' | unit= 'days' | unit='hour' | unit='hours' | unit='min' | unit='sec' | unit='ms')  
  ->  TIME_UNIT[$unit]
 ;

arithmeticRelation 
  : (rel='<' | rel='<=' | rel='=' | rel='!=' | rel='>' | rel='>=')
  ->  ARITHMETIC_RELATION[$rel]
  ;


/* Tokens */

ID
  : ('a'..'z' | 'A'..'Z') ('a'..'z' | 'A'..'Z' | '0'..'9' | '-' | '_')*
  ;

NUMBER
  : ('0'..'9')+ '.' ('0'..'9')+ Exponent? ('d' | 'f')?
  | ('0'..'9')+ Exponent ('d' | 'f')?
  | ('0'..'9')+ ('d' | 'f' | 'l')?
  ;

fragment
Exponent
  : ('e' | 'E') ('+' | '-')? ('0'..'9')+ 
  ;

STRING 
  : '"' (EscapeSequence | ~('\\' | '"'))* '"'
  ;

fragment
EscapeSequence
  : '\\' ('b' | 't' | 'n' | 'f' | 'r' | '\"' | '\'' | '\\')
  ;

COMMENT
  : '//' ~('\n' | '\r')* '\r'? '\n' {$channel=HIDDEN;}
  | '/*' (options {greedy=false;} : .)* '*/' {$channel=HIDDEN;}
  ;

WS
  : (' ' | '\t' | '\r' | '\n') {$channel=HIDDEN;}
  ;