package de.lmu.ifi.pms.cep.dura.compiler.dura.desugaring;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.IdentityHashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.ListIterator;
import java.util.Queue;

import org.antlr.runtime.ANTLRStringStream;
import org.antlr.runtime.CommonTokenStream;
import org.antlr.runtime.RecognitionException;
import org.antlr.runtime.tree.CommonTree;

import de.lmu.ifi.pms.cep.dura.compiler.antlr.DuraLexer;
import de.lmu.ifi.pms.cep.dura.compiler.antlr.DuraParser;


public class ConstantDefinitions extends Transformation {
	private ArrayList<CommonTree> implicitConstantDefinitions = new ArrayList<CommonTree>();
	private String[] implicitConstantDefinitionStrings = new String[] {
			
	};
	
	public ConstantDefinitions() {
		super();
		
		for (String defStr : implicitConstantDefinitionStrings) {
			ANTLRStringStream stream = new ANTLRStringStream(defStr);
			DuraLexer lexer = new DuraLexer(stream);
			CommonTokenStream tokens = new CommonTokenStream(lexer);
			DuraParser parser = new DuraParser(tokens);
			
			try {
				CommonTree term = (CommonTree) parser.constDefinition().getTree();
				
				normalizeTerm((CommonTree) term.getChild(0));
				
				implicitConstantDefinitions.add(term);
			} catch (RecognitionException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
		}
	}
	
	@Override
	@SuppressWarnings(value="unchecked")
	public void apply(CommonTree tree) {
		HashMap<String, CommonTree> nameToConstDefMap = new HashMap<String, CommonTree>();
		
		for (CommonTree typeDef : implicitConstantDefinitions) {
			// This should not happen.
			if (nameToConstDefMap.get(typeDef.getText()) != null) {
				throw new RuntimeException("Constant name \"" + typeDef.getText() + "\" is already bound to a different definition.");
			}
			
			nameToConstDefMap.put(typeDef.getText(), typeDef);
		}
		
		CommonTree searchRoot = (CommonTree) tree.getChild(0).getChild(0);
		if (searchRoot.getChildCount() > 0) {
			List<CommonTree> children = searchRoot.getChildren();
			for (CommonTree child : children) {
				if (child.getType() != DuraParser.CONSTDEF) {
					continue;
				}
				
				CommonTree temp = nameToConstDefMap.get(child.getText());
				if (temp != null) {
					// Let us allow type re-definitions (only) if they are equal.
					if (treesEqual(child, temp)) {
						continue;
					} else {
						throw new RuntimeException("Constant name \"" + child.getText() + "\" is already bound to a different (possibly implicit) definition (" + temp.toStringTree() + ").");
					}
				}
			
				nameToConstDefMap.put(child.getText(), child);
			}
			
			// Removes constant definitions from syntax tree!
			// children.clear();
			// searchRoot.freshenParentAndChildIndexes();
		}

		resolveInternalConstReferences(nameToConstDefMap);
		resolveExternalConstReferences(tree, nameToConstDefMap);
	}
	
	private void resolveExternalConstReferences(CommonTree tree, HashMap<String, CommonTree> nameToConstDefMap) {
		for (CommonTree constRef : collectConstantReferences(tree, nameToConstDefMap)) {
			CommonTree resolved = nameToConstDefMap.get(constRef.getChild(0).getText());
			CommonTree resolvedCopy = copyTree((CommonTree) resolved.getChild(0));
			
			int index = constRef.getChildIndex();
			constRef.getParent().replaceChildren(index, index, resolvedCopy);
		}
	}
	
	private void resolveInternalConstReferences(HashMap<String, CommonTree> nameToConstDefMap) {
		Collection<CommonTree> constDefs = nameToConstDefMap.values();
		
		// Time complexity estimates rely on the java implementation for LinkedList:
		// http://developer.classpath.org/doc/java/util/LinkedList-source.html
		final IdentityHashMap<CommonTree, LinkedList<CommonTree>> constDefToRefsMap =
			new IdentityHashMap<CommonTree, LinkedList<CommonTree>>();
		for (CommonTree constDef : constDefs) {
			constDefToRefsMap.put(constDef, collectConstantReferences(constDef, nameToConstDefMap));
		}
		
		// To speed things up the queue shall only contain constant definitions that
		// contain constant references, i.e. names referring to constants.
		Queue<CommonTree> constDefQueue = new LinkedList<CommonTree>();
		for (CommonTree constDef : constDefs) {
			if (constDefToRefsMap.get(constDef).size() > 0) {
				constDefQueue.add(constDef);
			}
		}
		
		int unexpandedConstDefCount = constDefQueue.size();
		int checkedConstDefCount = 0;
		
		while(!constDefQueue.isEmpty()) {
			CommonTree constDef = (CommonTree) constDefQueue.remove();
			
			boolean requeueConstDef = false;
			
			// We use an explicit list iterator for constant time removal.
			ListIterator<CommonTree> constRefsIter = constDefToRefsMap.get(constDef).listIterator(0);
			while (constRefsIter.hasNext()) {
				CommonTree constRef = constRefsIter.next();
				CommonTree resolved = nameToConstDefMap.get(constRef.getChild(0).getText());
				
				// The size attribute is incremented/decremented during linked list construction/destruction.
				// The method isEmpty() relies on size, which is retrievable in constant time.
				if (constDefToRefsMap.get(resolved).isEmpty()) {
					CommonTree resolvedCopy = copyTree((CommonTree) resolved.getChild(0));
					
					int index = constRef.getChildIndex();
					constRef.getParent().replaceChildren(index, index, resolvedCopy);
					
					// Remove this term paste from the linked list. This requires constant time.
					constRefsIter.remove();
				} else {
					requeueConstDef = true;
				}
			}
			
			if (requeueConstDef) {
				constDefQueue.add(constDef);
			}
			
			checkedConstDefCount++;
			if (checkedConstDefCount > unexpandedConstDefCount) {
				if (constDefQueue.size() == unexpandedConstDefCount) {
					StringBuilder sb = new StringBuilder();
					while (true) {
						CommonTree temp = (CommonTree) constDefQueue.remove();
						sb.append("\"");
						sb.append(temp.getText());
						sb.append("\"");
						
						if (constDefQueue.isEmpty()) {
							break;
						} else {
							sb.append(", ");
						}
					}
					
					throw new RuntimeException("Cycle in usage of constant name(s) " + sb.toString() + ".");
				} else {
					checkedConstDefCount = 0;
					unexpandedConstDefCount = constDefQueue.size();
				}
			}
		}
	}
	
	private LinkedList<CommonTree> collectConstantReferences(CommonTree tree, HashMap<String, CommonTree> nameToConstDefMap) {
		LinkedList<CommonTree> nodes = new LinkedList<CommonTree>();
		Queue<CommonTree> nodeQueue = new LinkedList<CommonTree>();
		nodeQueue.add(tree);
		
		while(!nodeQueue.isEmpty()) {
			CommonTree node = (CommonTree) nodeQueue.remove();			
			
			if (node.getType() == DuraParser.CONSTANT) {
				if (nameToConstDefMap.get(node.getChild(0).getText()) == null) {
					throw new RuntimeException("No constant definition for name " + node.getChild(0).getText() + " could be found.");
				}
				
				nodes.add(node);
				continue;
			}
			
			if (node.getChildCount() > 0) {
				for (Object c : node.getChildren()) {
					CommonTree child = (CommonTree) c;
					nodeQueue.add(child);
				}
			}
		}
		return nodes;
	}
}
