package de.lmu.ifi.pms.cep.dura.compiler.durac.queries;

import static de.lmu.ifi.pms.cep.dura.compiler.antlr.DuraCLexer.STATE_QUERY;
import static de.lmu.ifi.pms.cep.dura.compiler.durac.token_mappings.DuraDefaultAttributeNameMapping.eventIdentifierPrefix;
import static de.lmu.ifi.pms.cep.dura.compiler.durac.token_mappings.DuraDefaultAttributeNameMapping.eventIntervalName;
import static de.lmu.ifi.pms.cep.dura.compiler.durac.token_mappings.DuraDefaultAttributeNameMapping.identifierAttribute;
import static de.lmu.ifi.pms.cep.dura.compiler.durac.token_mappings.DuraDefaultAttributeNameMapping.identifierSeparator;
import static de.lmu.ifi.pms.cep.dura.compiler.durac.token_mappings.DuraDefaultAttributeNameMapping.intervalEndAttribute;
import static de.lmu.ifi.pms.cep.dura.compiler.durac.token_mappings.DuraDefaultAttributeNameMapping.labelSeparator;
import static de.lmu.ifi.pms.cep.dura.compiler.durac.token_mappings.DuraDefaultAttributeNameMapping.payloadLabel;
import static de.lmu.ifi.pms.cep.dura.compiler.durac.token_mappings.DuraDefaultAttributeNameMapping.soCreatedSuffix;
import static de.lmu.ifi.pms.cep.dura.compiler.durac.token_mappings.DuraDefaultAttributeNameMapping.soTerminatedSuffix;

import java.util.HashMap;
import java.util.Map;

import org.antlr.runtime.RecognitionException;
import org.antlr.runtime.tree.CommonTreeNodeStream;

import de.lmu.ifi.pms.cep.dura.compiler.antlr.StatefulObjectQueryTransformation;
import de.lmu.ifi.pms.cep.dura.compiler.antlr.utils.DuraImplicitTree;
import de.lmu.ifi.pms.cep.dura.compiler.antlr.utils.DuraTree;
import de.lmu.ifi.pms.cep.dura.compiler.durac.collections.DuraSchema;
import de.lmu.ifi.pms.cep.dura.compiler.durac.collections.Preamble;
import de.lmu.ifi.pms.cep.dura.compiler.durac.collections.QueryVariables;
import de.lmu.ifi.pms.cep.dura.compiler.durac.collections.TypeDefinitions;
import de.lmu.ifi.pms.cep.dura.compiler.durac.collections.UnambiguousQueryVariables;
import de.lmu.ifi.pms.cep.dura.compiler.durac.tsa.TimeInterval;
import de.lmu.ifi.pms.cep.dura.compiler.exceptions.DuraAmbiguousMappingException;
import de.lmu.ifi.pms.cep.dura.compiler.exceptions.DuraCompilationException;
import de.lmu.ifi.pms.cep.dura.compiler.exceptions.DuraTransformationException;
import de.lmu.ifi.pms.cep.dura.compiler.exceptions.DuraUnspecificCompliationException;
import de.lmu.ifi.pms.cep.tsa.TSA_SyntaxPatterns;
import de.lmu.ifi.pms.cep.tsa.conditions.BasicTypeAttribute;
import de.lmu.ifi.pms.cep.tsa.conditions.BasicTypeVariable;
import de.lmu.ifi.pms.cep.tsa.conditions.BinaryComparissonFormula;
import de.lmu.ifi.pms.cep.tsa.conditions.BinaryComparissonOperatorID;
import de.lmu.ifi.pms.cep.tsa.conditions.ConditionFormula;
import de.lmu.ifi.pms.cep.tsa.conditions.NamedTypeVariable;
import de.lmu.ifi.pms.cep.tsa.conditions.Variable;
import de.lmu.ifi.pms.cep.tsa.schema.BasicTypeID;
import de.lmu.ifi.pms.cep.tsa.schema.TemporalStreamSchemaTypeID;

public class StatefulObjectQuery {
	public final AtomicEventQuery positiveQuery;
	public final AtomicEventQuery negativeQuery;

	public final UnambiguousQueryVariables variables; 
	
	public final boolean queriesStaticObject;
	public final ConditionFormula joinCondition;
	public final Map<String,String> identifierMappig = new HashMap<String, String>();
	
	public StatefulObjectQuery(DuraTree query, final Preamble preamble, final DuraSchema schema) {
		assert(query.getType() == STATE_QUERY);
		
		DuraTree queryIdentifier = query.getChild(0);
		String queryIdentifierName = queryIdentifier.getChild(0).getText();

		String createdEventIdentifierName = String.format("%s%s%s", queryIdentifierName, identifierSeparator, soCreatedSuffix);
		String terminatedEventIdentifierName = String.format("%s%s%s", queryIdentifierName, identifierSeparator, soTerminatedSuffix);
		
		String createdEventIdentifier = String.format("%s %s", eventIdentifierPrefix, createdEventIdentifierName);
		String terminatedEventIdentifier = String.format("%s %s", eventIdentifierPrefix, terminatedEventIdentifierName);

		Map<String,String> substitutePrefix = new HashMap<String, String>();
		substitutePrefix.put(queryIdentifierName, TSA_SyntaxPatterns.attributeName(createdEventIdentifierName, payloadLabel));		
		
		try {
			negativeQuery = new AtomicEventQuery(new DuraImplicitTree(new StatefulObjectQueryTransformation(new CommonTreeNodeStream(query)).terminatedQuery().getTree(), query), substitutePrefix, preamble, schema);
			positiveQuery = new AtomicEventQuery(new DuraImplicitTree(new StatefulObjectQueryTransformation(new CommonTreeNodeStream(query)).createdQuery().getTree(), query), substitutePrefix, preamble, schema);
		} catch (RecognitionException e) {
			throw new DuraTransformationException(e, query);
		}
		
		try {
			queriesStaticObject = schema.streams.getStreamByName(TemporalStreamSchemaTypeID.EVENT, preamble, String.format("%s%s%s", query.getChild(1).getChild(0).getText(), labelSeparator, soCreatedSuffix)).isStatic;
		} catch (DuraUnspecificCompliationException e) {
			throw new DuraCompilationException(e, query);
		}
			
		//add mapping for state s to according event id and timestamps
		BasicTypeVariable idVar = (BasicTypeVariable) QueryVariables.variableFromToken(queryIdentifier);
		Variable<?> intervalVar = NamedTypeVariable.create(idVar.name, TypeDefinitions.TIME_INTERVAL);
		
		BasicTypeAttribute idAttr = BasicTypeAttribute.create(TSA_SyntaxPatterns.attributeName(createdEventIdentifierName, payloadLabel, identifierAttribute), BasicTypeID.IDENTIFIER.type);
		BasicTypeAttribute beginTs = BasicTypeAttribute.create(TSA_SyntaxPatterns.attributeName(createdEventIdentifierName, eventIntervalName, intervalEndAttribute), BasicTypeID.TIMESTAMP.type);
		BasicTypeAttribute endTs = BasicTypeAttribute.create(TSA_SyntaxPatterns.attributeName(terminatedEventIdentifierName, eventIntervalName, intervalEndAttribute), BasicTypeID.TIMESTAMP.type);
		
		UnambiguousQueryVariables idMapping = new UnambiguousQueryVariables(idVar, idAttr);
		UnambiguousQueryVariables intervalMapping = new UnambiguousQueryVariables(intervalVar, new TimeInterval(beginTs, endTs));
		
		try {
			variables = QueryVariables.unionUnambiguousMappings(idMapping, intervalMapping, positiveQuery.getVariables().dropMappings(createdEventIdentifier), negativeQuery.getVariables().dropMappings(terminatedEventIdentifier)).convertToUnambiguousMapping();
		} catch (DuraAmbiguousMappingException e) {
			throw new DuraCompilationException(e, query);
		}
		
		//create join condition for positive and negative query
		BasicTypeAttribute createdId = BasicTypeAttribute.create(TSA_SyntaxPatterns.attributeName(createdEventIdentifierName, payloadLabel, identifierAttribute), BasicTypeID.IDENTIFIER);
		BasicTypeAttribute terminatedId = BasicTypeAttribute.create(TSA_SyntaxPatterns.attributeName(terminatedEventIdentifierName, payloadLabel, identifierAttribute), BasicTypeID.IDENTIFIER);
		
		joinCondition = new BinaryComparissonFormula(BinaryComparissonOperatorID.EQUAL, createdId, terminatedId);
		
		//add mapping of primal state identifier to identifier of the create event
		identifierMappig.put(queryIdentifierName, TSA_SyntaxPatterns.attributeName(createdEventIdentifierName, payloadLabel));
	}
}
