package de.lmu.ifi.pms.cep.event_mill.io.connect;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import de.lmu.ifi.pms.cep.event_mill.io.stream.SinkStream;
import de.lmu.ifi.pms.cep.event_mill.io.stream.SourceStream;
import de.lmu.ifi.pms.cep.event_mill.io.stream.StreamIO;
import de.lmu.ifi.pms.cep.tsa.schema.BasicTypeID;

import de.lmu.ifi.pms.java.util.collection.array.fixed.FixedArrayNavigableSet;

public class StreamConnector extends Connector {
	
	protected final SourceStream[] sources;
	protected final SinkStream[] sinks;
	protected final int[] numberOfAttributess;
	protected final BasicTypeID[][] typess;
	
	public StreamConnector(final StreamIO io1, final StreamIO io2) {
		FixedArrayNavigableSet<String> commonNames = FixedArrayNavigableSet.intersect(io1.sourceNames(), io2.sinkNames());
		this.sources = new SourceStream[commonNames.size()];
		this.sinks = new SinkStream[this.sources.length];
		this.numberOfAttributess = new int[this.sources.length];
		this.typess = new BasicTypeID[this.sources.length][];
		SourceStream source;
		SinkStream sink;
		int numberOfAttributes;
		BasicTypeID[] types;
		for(int messageType = 0; messageType < this.sources.length; ++messageType){
			source = io1.sourceStream(commonNames.get(messageType));
			this.sources[messageType] = source;
			sink = io2.sinkStream(commonNames.get(messageType));
			this.sinks[messageType] = sink;
			assert sink.name.equals(source.name);
			
			numberOfAttributes = source.attributes.size();
			this.numberOfAttributess[messageType] = numberOfAttributes;
			types = new BasicTypeID[numberOfAttributes];
			this.typess[messageType] = types;
			for(int attribute = 0; attribute < numberOfAttributes; ++attribute){
				types[source.position(attribute)] = source.types.values().get(attribute);
			}
			
			//check compatibility of types
			for(int attribute = 0; attribute < numberOfAttributes; ++attribute){
				if(!sink.types.keys().get(attribute).equals(source.types.keys().get(attribute))){
					throw new IllegalArgumentException("Type \"" + source.name + "\" has different attributes for source and sink stream.");
				}
				if(sink.types.values().get(attribute) != source.types.values().get(attribute)){
					throw new IllegalArgumentException("Attribute \"" + sink.types.keys().get(attribute) + "\"of type \"" + source.name + "\" has a different type for source and sink stream.");
				}
			}
		}
	}
	

	@Override
	public void nextBlock() {
		SourceStream source;
		SinkStream sink;
		int numberOfAttributes;
		BasicTypeID[] types;
		for(int messageType = 0; messageType < this.sources.length; ++messageType){
			source = this.sources[messageType];
			sink = this.sinks[messageType];
			numberOfAttributes = this.numberOfAttributess[messageType];
			types = this.typess[messageType];
			assert source.name.equals(sink.name);
			if(source.nextBlock()){
				while(source.next()){
					boolean nextSucceded = sink.next();
					if(!nextSucceded){
						sink.nextBlock();
						nextSucceded = sink.next();
						assert nextSucceded;
					}
					BasicTypeID attributeType;
			    	for (int attribute = 1; attribute <= numberOfAttributes; ++attribute){
			    		attributeType = types[attribute - 1];
			    		switch (attributeType) {
						case BOOLEAN:
							sink.setBoolean(attribute, source.getBoolean(attribute));
							break;
						case STRING:
							sink.setString(attribute, source.getString(attribute));
							break;
						case INT:
							sink.setInt(attribute, source.getInt(attribute));
							break;
						case LONG:
							sink.setLong(attribute, source.getLong(attribute));
							break;				
						case FLOAT:
							sink.setFloat(attribute, source.getFloat(attribute));
							break;
						case DOUBLE:
							sink.setDouble(attribute, source.getDouble(attribute));
							break;
						case TIMESTAMP:
							sink.setTimestamp(attribute, source.getTimestamp(attribute));
							break;				
						case DURATION:
							sink.setDuration(attribute, source.getDuration(attribute));
							break;				
						case IDENTIFIER:
							sink.setIdentifier(attribute, source.getIdentifier(attribute));
							break;
						case REFERENCE: throw new IllegalArgumentException("Type \"reference\" not supported.");
						default:
							break;
						}
			    	}
				}
				sink.nextBlock();
			}
		}
	}

	protected static final Logger logger = LoggerFactory.getLogger(StreamConnector.class);
	
	@Override
	protected Logger logger(){
		return logger;
	}
}
